package Stanford::Orange::LogTrend ;

use strict ;
use warnings ;
use autodie ;

use base qw( Exporter ) ;
our @EXPORT_OK = qw
  (
  delete_old_files
  find_run_in_files
  find_run_in_files_gzipped
  format_trend_results
  get_archive_directory
  list_files_in_order
  ) ;

use Data::Dumper ;
use Carp ;
use File::Spec ;

# You can supply a reference to a Perl function that will be run on each
# file and return a new file name in case you want to do log file
# preprocessing. An example of this pre-processing might be to run the log
# files through filter-syslog.
sub find_run_in_files {
    my ($files_aref, $pre_process_file_fref) = @_ ;

    my @files       = @{$files_aref} ;
    my @FILE_HASHES = () ;

    # If no pre-process function reference is passed in, use the identity
    # function.
    if (!$pre_process_file_fref) {
        $pre_process_file_fref = sub {
            my ($file) = @_ ;
            return $file ;
        } ;
    }

    foreach my $file (@files) {
        my $file_to_process = $pre_process_file_fref->($file) ;

        # We store the lines from this file that appear in ALL previous
        # files in @file_lines
        my @file_lines = () ;

        open(my $FH, '<', $file_to_process) ; ## no critic (RequireBriefOpen)
        my $lines_href = {} ;
        while (my $line = <$FH>) {
            chomp $line ;

            # Skip some lines:
            #  1. Compiled catalog for www04.stanford.edu in environment ...
            if ($line =~ m{Compiled[ ]catalog[ ]for[ ]\S*[ ]in[ ]environment}xsm) {
                next ;
            #  2. Not sending tagmail report; no changes
            } elsif ($line =~ m{Not[ ]sending[ ]tagmail[ ]report;[ ]no[ ]changes}xsm) {
                next ;
            } elsif ($line =~ m{Starting\s+Puppet\s+master\s+version}xsm) {
                next ;
            } else {

                # May 30 23:45:18 jimhenson1 puppet-master[7622]:
                $line =~ s{^.*puppet-master\[\d+\]:\s*}{}xsm ;
                if (!exists($lines_href->{$line})) {
                    $lines_href->{$line} = 1 ;
                    push(@file_lines, $line) ;
                }
            }
        }

        push(@FILE_HASHES, [ $lines_href, \@file_lines ]) ;

        close($FH) ;
    }

    my ($top_lines_href, $top_file_lines_aref) = @{ $FILE_HASHES[0] } ;
    my %line_to_count = () ;
    my @counts        = () ;
    foreach my $line (@{$top_file_lines_aref}) {
        my $count = calculate_line_run_length($line, 0, \@FILE_HASHES) ;
        $line_to_count{$line} = $count ;

        if ($counts[$count]) {
            my $all_lines_with_this_count_href = $counts[$count] ;
            push(@{$all_lines_with_this_count_href}, $line) ;
        } else {
            $counts[$count] = [$line] ;
        }
    }

    return \@counts ;
}

# Calculate the longest run of files that this line appears in, starting
# at the file in position $position.
sub calculate_line_run_length {
    my ($line, $position, $file_hashes_aref) = @_ ;

    my @FILE_HASHES = @{$file_hashes_aref} ;

    # If $position extends beyond the end of $file_hashes_aref, then
    # simply return 0.
    if ($position >= scalar(@FILE_HASHES)) {
        return 0 ;
    }

    # Get the top element of @FILES and see if that element has $line in
    # it. If yes, then return 1 + calculate_line_run_length($line)
    my ($top_lines_href, $top_file_lines_aref) = @{ $FILE_HASHES[$position] } ;
    if ($top_lines_href->{$line}) {
        return (1 +
              calculate_line_run_length($line, $position + 1, $file_hashes_aref)
        ) ;
    } else {
        return 0 ;
    }
}

# Make sure there are at most N files in a directory whose file names
# match a regular expression.
#
# Given a directory, a regular expression, and a number N, do the
# following:
#
# 1. Construct a list of files in the provided directory whose names match
# the provided resular expression. This list will be in order of file
# modify date with the newest file first in the list.
#
# 2. If the number of files in the list is M and M > N, then
# take the last M - N files from the list and delete them from the
# provided directory.
#
# Returns the array of files deleted.
sub delete_old_files {
    my ($dir, $regex, $at_most) = @_ ;

    if (!$dir) {
        croak 'missing required directory parameter' ;
    }

    if (!$regex) {
        croak 'missing required regular expression parameter' ;
    }

    if (!defined($at_most)) {
        croak 'missing required number of files to keep parameter' ;
    } elsif ($at_most <= 0) {
        croak 'the required number of files to keep parameter must be positive' ;
    }

    my @files_in_order = list_files_in_order($dir, $regex) ;

    my $M = scalar(@files_in_order) ;

    my @files_to_delete = ()  ;
    if ($M > $at_most) {
        @files_to_delete = @files_in_order[$at_most  .. ($M - 1)] ;
    }

    foreach my $file (@files_to_delete) {
        unlink $file ;
    }

    return @files_to_delete ;
}

# Get an AFS archive directory path.
#
# Let A be a directory (usually in AFS), F the full path to a file.
#
# 1. Let X be the last-modified date of F.
# 2. Let YYYY = year(X), MM = month(X), DD = day(X).
# 2. Return the string "A/YYYY/MM/YYYY-MM-DD

sub get_archive_directory {
    my ($dir, $file) = @_ ;

    my $mtime = (stat $file)[9];

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
        localtime($mtime);


    my $YYYY = 1900 + $year ;
    my $MM   = $mon  + 1 ;
    my $DD   = $mday ;

    if ($MM < 10) {
        $MM = '0' . $MM ;
    }
    if ($DD < 10) {
        $DD = '0' . $DD ;
    }

    $dir =~ s{\/$}{}xsm ;

    return "$dir/$YYYY/$MM/$YYYY-$MM-$DD" ;
}

sub format_trend_results {
    my ($counts_aref) = @_ ;
    my $rv = q{} ;

    my $separator = q{-} x 20 ;

    my @counts  = reverse @{$counts_aref} ;
    my $counter = scalar(@counts) ;
    foreach my $lines_aref (@counts) {
        --$counter ;
        if (!$lines_aref) {
            next ;
        }

        my $s ;
        if ($counter == 1) {
            $s = q{} ;
        } else {
            $s = 's' ;
        }

        $rv .= "Appears in $counter consecutive day$s\n" ;
        $rv .= $separator . "\n" ;

        my @lines = @{$lines_aref} ;
        foreach my $line (@lines) {
            $rv .= $line . "\n" ;
        }
        $rv .= "\n" ;
    }

    return $rv ;
}

sub find_run_in_files_gzipped {
    my ($files_aref) = @_ ;
    return find_run_in_files($files_aref, \&preprocess_gzipped) ;
}

sub preprocess_gzipped {
    my ($file) = @_ ;

    use File::Temp ;
    my ($fh, $tempfile) = File::Temp::tempfile('logtrendXXXXX', UNLINK => 1) ;

    use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
    gunzip $file => $tempfile
      or croak "gunzip failed: $GunzipError\n" ;

    return $tempfile ;
}

# Given a directory and a regular expression, return an array of full path
# names to files in that directory that match that regular expression.
# The array of file paths are returned in order of last-modified date,
# with the NEWEST file listed FIRST.
sub list_files_in_order {
    my ($directory, $regex) = @_ ;

    opendir(my $DIR, $directory) ;

    my @files = () ;
    while (my $filename = readdir($DIR)) {
        my $curfile = File::Spec->catfile(($directory), $filename) ;

        # We only want files
        if (!(-f $curfile)) {
            next ;
        }

        # Does the filename match?
        if ($filename =~ m{$regex}sm) {  ## no critic (RequireExtendedFormatting)
            push(@files, [ $curfile, (stat($curfile))[9] ]) ;
        }
    }

    closedir($DIR) ;

    my $sort_aux = sub {
        return ($b->[1] <=> $a->[1]) ;
    } ;

    my @files_sorted = sort $sort_aux @files ;

    # Get the first element.
    @files_sorted = map { $_->[0] } @files_sorted ;

    return @files_sorted ;
}

1 ;

__END__

=for stopwords pathnames GZIP Lewenberg perldoc

=head1 NAME

Stanford::Orange::LogTrend - Look for recurring lines in a series of files

=head1 VERSION

Version 0.01

=head1 SYNOPSIS

    use Stanford::Orange::LogTrend qw/find_run_in_files format_trend_results/ ;

    my $results_aref = find_run_in_files(\@files) ;

    # Prints out the results in a nice, human-readable format.
    print format_trend_results($results_aref) ;

=head1 DESCRIPTION

Given an ordered set of log files, for each line in the first file find how
many of the succeeding files have that same line (ignores the initial date
string). This gives an indication of log message trends.

=head1 SUBROUTINES/METHODS

=head2 C<find_run_in_files>

Takes as input a reference to an array of file pathnames. The return value
is an array, where the Nth element of the array contains an array
reference to a list of lines of the first file that appear in the first N
files.

Here is an example. If C<file1.txt> contains

    A
    B
    C
    D

the file C<file2.txt> contains

    A
    B
    D

and the file C<file3.txt> contains

    A
    B
    C

Then, a call to

    find_run_in_file(['file1.txt', 'file2.txt', 'file3.txt'])

returns this data structure

    [
      undef,
      ['C'],
      ['D'],
      ['A', 'B']
    ]


=head2 C<find_run_in_files_gzipped>

This function works in exactly the same way as C<find_run_in_files> except
it expects the files passed in to be in GZIP format.

=head2 C<format_trend_results>

Given the output C<find_run_in_files> returns a human-readable string.
For example, using the same three files as in the above example:

    my $results_aref = find_run_in_file(['file1.txt', 'file2.txt', 'file3.txt']) ;
    print format_trend_results($results_aref) ;

would print

    Appears in 3 consecutive days
    --------------------
    A
    B

    Appears in 2 consecutive days
    --------------------
    D

    Appears in 1 consecutive days
    --------------------
    C

=head2 C<list_files_in_order>

Given a directory and a regular expression, return an array of path
names to files in that directory that match that regular expression in
order of most recent first. For example, if C</tmp/> contains these files:

    10 Jun  3 08:20 filea.txt
    12 Jun  3 08:01 file0.txt
    16 Jun  3 09:04 fileb.txt

Then

    list_files_in_order('/tmp', '^file..txt$')

will return the array

    ('/tmp/fileb.txt', '/tmp/filea.txt', '/tmp/file0.txt')

=head1 DIAGNOSTICS

Later

=head1 CONFIGURATION AND ENVIRONMENT

None needed.

=head1 DEPENDENCIES

Carp, File::Spec

=head1 INCOMPATIBILITIES

None known.

=head1 BUGS AND LIMITATIONS

None known.

=head1 AUTHOR

Adam Lewenberg, C<< <adamhl at stanford.edu> >>

=head1 BUGS

Please report any bugs or feature requests
to C<adamhl at stanford.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Stanford::LogTrend

=head1 LICENSE AND COPYRIGHT

Copyright 2013 Stanford University, all rights reserved.

=cut
