package Stanford::Orange::Nagios ;

use warnings ;
use strict ;
use autodie ;

## no critic (RequireArgUnpacking, RequireCarping, RequireCheckedSyscalls) ;

# Make this a Moose class
use Moose ;
use namespace::autoclean ;

my $EXIT_STATE_TEXT_TO_VALUE = {
    'OK'       => 0,
    'WARNING'  => 1,
    'CRITICAL' => 2,
    'UNKNOWN'  => 3,
} ;

has warning_level => (
    is  => 'rw',
    isa => 'Str',
) ;

has critical_level => (
    is  => 'rw',
    isa => 'Str',
) ;

has exit_state_text => (
    is      => 'rw',
    isa     => 'Str',
    default => q{},
) ;

has exit_message => (
    is      => 'rw',
    isa     => 'Str',
    default => q{},
) ;

sub set_critical {
    my $self = shift ;

    $self->exit_state_text('CRITICAL') ;
    return ;
}

sub set_warning {
    my $self = shift ;

    $self->exit_state_text('WARNING') ;
    return ;
}

sub set_unknown {
    my $self = shift ;

    $self->exit_state_text('UNKNOWN') ;
    return ;
}

sub set_ok {
    my $self = shift ;

    $self->exit_state_text('OK') ;
    return ;
}

sub set_exit_message {
    my $self = shift ;
    my ($message) = @_ ;

    $self->exit_message($message) ;
    return ;
}

# Once the exit_message and exit_state_text are set, we can print out the
# message and exit with the proper return code.
sub finish {
    my $self = shift ;

    if (!$self->exit_state_text) {
        die 'you must set an exit state (ok, warning, critical, or '
          . 'unknown) before finishing' ;
    }

    if (!$self->exit_message) {
        die 'you must call set_exit_message before finishing' ;
    }

    print $self->exit_message . "\n" ;
    exit($EXIT_STATE_TEXT_TO_VALUE->{ $self->exit_state_text }) ;
}

sub to_s {
    my $self = shift ;

    my $exit_message    = $self->exit_message ;
    my $exit_state_text = $self->exit_state_text ;

    my $not_defined = '<NOT DEFINED>' ;

    if ($exit_message eq q{}) {
        $exit_message = $not_defined ;
    }

    my $rv = $not_defined ;
    if ($exit_state_text eq q{}) {
        $exit_state_text = $not_defined ;
    } else {
        $rv = $EXIT_STATE_TEXT_TO_VALUE->{$exit_state_text} ;
    }

    return <<"EOS";
exit_code:       $rv
exit_state_text: $exit_state_text
exit_message:    $exit_message
EOS

}
no Moose ;
__PACKAGE__->meta->make_immutable ;

1 ;

__END__

=for stopwords nagios Lewenberg perldoc

=head1 NAME

Stanford::Orange::Nagios - Nagios script helper

=head1 VERSION

Version 0.01

=head1 SYNOPSIS

    use Stanford::Orange::Nagios ;

    # ...your own code that tests something...

    my $nagios = Stanford::Orange::Nagios->new() ;

    if ( something bad ) {
        $nagios->set_critical() ;
        $nagios->set_exit_message('system is critically bad') ;
    } elsif ( something not so bad ) {
        $nagios->set_warning() ;
        $nagios->set_exit_message('system is getting bad') ;
    } elsif ( not sure what is going on ) {
        $nagios->set_unknown() ;
        $nagios->set_exit_message('system is in an undefined state') ;
    } else ( everything's OK ) {
        $nagios->set_ok() ;
        $nagios->set_exit_message('system is doing fine') ;
    }

    # This should be the LAST line of your Nagios script
    $nagios->finish() ;

=head1 DESCRIPTION

A helper class to make writing Nagios scripts easier.

Nagios expects an exit value of 0 (OK), 1 (WARNING), 2 (CRITICAL), or 3
(UNKNOWN). This helper class hides most of the detail, so all you have to
do is instantiate a nagios object with C<new>, set an exit state and
message, and then call the C<finish> method.

To set an exit state, call one of these methods:

    set_ok()
    set_warning()
    set_critical()
    set_unknown()

To set the exit message use the C<set_exit_message> method:

    set_exit_message('this is the exit message')

Finally, call the C<finish> method when you are all done.

It is an error to call the C<finish> if you have not set the exit state or
an exit message.

=head1 SUBROUTINES/METHODS

See above.

=head1 CONFIGURATION AND ENVIRONMENT

None needed.

=head1 DIAGNOSTICS

None needed.

=head1 DEPENDENCIES

Moose

=head1 INCOMPATIBILITIES

None known.

=head1 BUGS AND LIMITATIONS

None known.

=head1 AUTHOR

Adam Lewenberg, C<< <adamhl at stanford.edu> >>

=head1 BUGS

Please report any bugs or feature requests
to C<adamhl at stanford.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Stanford::Orange::Nagios

=head1 LICENSE AND COPYRIGHT

Copyright 2015 Stanford University, all rights reserved.

=cut




