package Stanford::Orange::Util ;

## no critic (InputOutput::RequireCheckedSyscalls);

use warnings ;
use strict ;
use autodie ;

use base qw( Exporter ) ;
our @EXPORT_OK = qw
  (
  config_file_lines
  dn_to_cn
  extract_pod
  print_man_page_from_pod
  files_older_than
  get_wallet
  kerberos_ticket_cache
  kerberos_ticket_cache_no_afs
  mysql_time
  number_recent_files
  run_command
  run_command_improved
  run_under_keytab
  slurp
  spew
  tail_file_matches
  tempfile
  trim
  ) ;

our $VERSION = '0.09' ;

use Carp ;
use File::Spec ;
use File::Temp ;
use IPC::Run qw/ run/ ;

# __COVER__ uncoverable_file tests/.uncoverable

sub tempfile {
    ## no critic (ProhibitStringyEval, RequireCheckingReturnValueOfEval);
    eval 'use File::Temp' ;
    ## use critic
    my ($fh, $filename) = File::Temp::tempfile() ;

    close($fh) ;  # Uses autodie.

    return $filename ;
}

sub slurp {
    my ($filename, $filesize_limit) = @_ ;

    # If filesize is defined but is non-positive, it is silently ignored.
    if (defined($filesize_limit) && (($filesize_limit + 0) > 0) && (-e $filename))
    {

        # Get the file size (in bytes) of $filename
        my $filesize = -s $filename ;
        if ($filesize > $filesize_limit) {
            croak "could not open file '$filename' for slurping: "
              . "file '$filename' has $filesize bytes exceeding "
              . "the limit of $filesize_limit" ;
        }
    }

    my $text = q{} ;
    {
        ## no critic (Variables, InputOutput)

        local ($/, *FH) ;

        # Turn off autodie for a bit.
        {
            no autodie qw(open) ;
            if (!open(FH, '<', $filename)) {
                croak "can't open file $filename for slurping: $!" ;
            }
        }

        $text = <FH> ;
        ## use critic
    }

    return $text ;
}

sub spew {
    my ($x, $filename) = @_ ;

    open(my $FH, '>', $filename) ;

    ## no critic (InputOutput::RequireCheckedSyscalls);
    print {$FH} $x ;

    close($FH) ;

    return 1 ;
}

sub trim {
    my ($s) = @_ ;

    if (!defined($s)) { return $s }

    $s =~ s{^\s*}{}xs ;  ## no critic (RequireLineBoundaryMatching) ;
    $s =~ s{\s*$}{}xs ;  ## no critic (RequireLineBoundaryMatching) ;

    return $s ;
}

sub run_command {
    my ($cmd) = @_ ;

    my $temp_stdout = tempfile() ;
    my $temp_stderr = tempfile() ;

    ## no critic (ProhibitBacktickOperators)
    my $rv = `$cmd 1>$temp_stdout 2>$temp_stderr` ;
    ## use critic

    my $exit_value = $? >> 8 ;

    my $stdout = slurp($temp_stdout) ;
    my $stderr = slurp($temp_stderr) ;

    # Because we have autodie turned on, these will die
    # if they cannot unlink.
    unlink($temp_stdout) ;
    unlink($temp_stderr) ;

    return ($stdout, $stderr, $exit_value) ;
}

# Use: pass in an array, returns ($stdout, $stderr, $exit_value)
sub run_command_improved {
    my (@command) = @_ ;

    # (From Russ)
    my ($out, $err) ;
    IPC::Run::run(\@command, q{>}, \$out, q{2>}, \$err) ;

    return ($out, $err, $? >> 8) ;
}

sub mysql_time {

    # # ##  # # ##  # # ##  # # ##  # # ##  # # ##  # # ##
    my $pad_fref = sub {
        my ($n) = @_ ;
        if   ($n + 0 < 10) { return '0' . $n ; }
        else               { return q{} . $n ; }
    } ;  ## ## end $pad_fref sub

    my @t = localtime() ;

    #<<< tell perltidy to skip this small section
    my $year  =             $t[5] + 1900 ;
    my $month = $pad_fref->($t[4] + 1) ;
    my $day   = $pad_fref->($t[3]) ;
    my $hour  = $pad_fref->($t[2]) ;
    my $min   = $pad_fref->($t[1]) ;
    my $sec   = $pad_fref->($t[0]) ;
    #>>>
    return qq{$year-$month-$day $hour:$min:$sec} ;
}

# Converts a text file into an array with each line being an element of
# the array. It remove any comments and any blank lines. All lines added
# to the array are trimmed.
#
# Example: If the file "config.txt" contains these five lines:
#
# # This is a config file.
# abc 123 xyz
#   Hello, world!
#
# last line of the file
#
# The result of config_file_lines('config.txt') will be the array
#
# ('abc 123 xyz', 'Hello, world!', 'last line of the file')
sub config_file_lines {
    # uncoverable subroutine
    my ($file) = @_ ;

    my $file_as_string = slurp($file) ;

    ## no critic (RegularExpressions)
    my @lines_raw = split(/\n/, $file_as_string) ;
    ## use critic

    my @lines = () ;
    foreach my $line (@lines_raw) {
        if ($line =~ m{^\s*\#}xsm) {

            # Skip comments.
            next ;
        }

        $line = trim($line) ;

        if ($line eq q{}) {

            # Skip blank lines
            next ;
        }

        push(@lines, $line) ;
    }

    return @lines ;
}

# Given a command array and kerberos keytab file, run the command under
# the context of that keytab.
#
# Returns the triple ($stdout, $stderr, $exit_code) from running the
# k5start command.  Note that we include the -t option to get an AFS
# token.
sub run_under_keytab {
    # uncoverable subroutine
    my ($command_aref, $keytab_file) = @_ ;

    my @command_to_run = @{$command_aref} ;

    my @k5start_cmd = (
        'k5start',
        '-U',
        '-f', $keytab_file,
        '-t',  # get AFS token
        '-q',  # be quiet
    ) ;

    #<<< tell perltidy to skip this small section
    my @command = (@k5start_cmd,
                   q{--},
                   'sh',
                   '-c',
                   join(q{ }, @command_to_run),
        ) ;
    #>>>
    return run_command_improved(@command) ;
}


# Create a Kerberos ticket cache and point KRB5CCNAME at it.
# Requires a keytab file. Optional third argument is the principal
# name. Note that this function attempts to get an AFS token. If you don't
# want to provision an AFS token call kerberos_ticket_cache_no_afs instead.
sub kerberos_ticket_cache {
    my ($keytab_file, $cache_location, $principal) = @_ ;

    # The last parameter is 1 to indicate we want an AFS token.
    return kerberos_ticket_cache_aux($keytab_file, $cache_location, $principal, 1) ;
}

sub kerberos_ticket_cache_no_afs {
    my ($keytab_file, $cache_location, $principal) = @_ ;

    # The last parameter is 0 to indicate we do NOT want an AFS token.
    return kerberos_ticket_cache_aux($keytab_file, $cache_location, $principal, 0) ;
}

# This is the main function used by the wrapper functions
# kerberos_ticket_cache and kerberos_ticket_cache_no_afs.
sub kerberos_ticket_cache_aux {
    my ($keytab_file, $cache_location, $principal, $get_afs_token) = @_ ;

    my @options = () ;

    # If no principal add the '-U' option.
    if (!$principal) {
        push(@options, '-U') ;
    }

    # Add keytab_file, cache_location, and quiet options.
    push(@options, '-f', $keytab_file) ;
    push(@options, '-k', $cache_location) ;
    push(@options, '-q') ;

    # If $get_afs_token is true, add the get AFS token option.
    if ($get_afs_token) {
	push(@options, '-t') ;
    }

    # Combine the command and the options.
    my @cmd = ('k5start', @options) ;

    # If $principal was supplied make it the last argument.
    if ($principal) {
        push(@cmd, $principal) ;
    }

    my ($stdout, $stderr, $exit_status) = run_command_improved(@cmd) ;

    # Set the KRB5CCNAME environment variable. Do NOT use local here as
    # that would mean KRB5CCNAME environment variable would lose its value
    # as soon as this function returns.
    ## no critic (Variables::RequireLocalizedPunctuationVars);
    $ENV{'KRB5CCNAME'} = $cache_location ;

    return 1;
}


# Given a directory, find the number of files in that
# directory that are have a last-modified date no older than the provided
# number of seconds. Only looks at regular files and does NOT recurse into
# subdirectories.
#
# An optional thrid argument does a regular expression match on the
# _entire_ file name (basename plus extension but not the path). Only
# files matching that regular expression are counted. This match is
# case-sensitive.
sub number_recent_files {
    my ($dir, $age_limit_in_seconds, $file_rx) = @_ ;

    opendir(my $DIR, $dir) ;

    # If a file's last-modifided date is before $age_limit_epoch, then
    # this file does not get counted.
    my $age_limit_epoch = time() - $age_limit_in_seconds ;

    my $counter = 0 ;
    while (my $file = readdir($DIR)) {
        my $fullpath = File::Spec->catdir(($dir), $file) ;

        # Skip if $fullpath is not a regular file
        if (!(-f $fullpath)) {
            next ;
        }

        # If $file_rx is defined, match against it.
        if ($file_rx && ($file !~ m{$file_rx}xsm)) {

            # This file does not match, so skip.
            next ;
        }

        # Get the last-modified date.
        my $last_modified_epoch = (stat($fullpath))[9] ;

        # Is this last-modified-date no older than our limit? If so,
        # increment our counter.
        if ($last_modified_epoch >= $age_limit_epoch) {
            $counter = $counter + 1 ;
        }
    }

    closedir($DIR) ;

    return $counter ;
}

# Given a directory, return a list of of files (or directories) in that
# directory that have a last modified date older than the provided
# date. Does not return '.' or '..'. The items in the list are full paths.
#
# An optional thrid argument does a regular expression match on the
# _entire_ file name (basename plus extension but not the path). Only
# files/directories matching that regular expression are counted. This
# match is case-sensitive.
sub files_older_than {
    my ($dir, $age_limit_in_seconds, $file_rx) = @_ ;

    if (!(defined $dir)) {
        croak 'files_older_than: missing required directory parameter' ;
    }

    if (!(defined $age_limit_in_seconds)) {
        croak 'files_older_than: missing required age-limit parameter' ;
    }

    opendir(my $DIR, $dir) ;

    # If the file's last-modified date is AFTER $age_limit_epoch, then
    # we skip it.
    my $age_limit_epoch = time() - $age_limit_in_seconds ;

    my @found = () ;
    while (my $file = readdir($DIR)) {

        # Skip if $file is '.' or '..'
        if (($file eq q{.}) || ($file eq q{..})) {
            next ;
        }

        # If $file_rx is defined, match against it.
        if ($file_rx && ($file !~ m{$file_rx}xsm)) {

            # This file does not match, so skip.
            next ;
        }

        my $fullpath = File::Spec->catdir(($dir), $file) ;

        # Get the last-modified date.
        my $last_modified_epoch = (stat($fullpath))[9] ;

        # Is this last-modified-date no older than our limit? If so,
        # skip.
        if ($last_modified_epoch >= $age_limit_epoch) {
            next ;
        }

        # If we get here, then $fullpath is older then the supplied date,
        # so add it to our hit list.
        push(@found, $fullpath) ;
    }

    return @found ;
}

# Extract the CN from a DN.
#
# Example:
# '/C=US/ST=California/L=Stanford/O=Stanford University/OU=IT Services MDM/CN=ios-device-test624.stanford.edu'
# returns 'ios-device-test624.stanford.edu',
#
# If the passed in DN is empty or undefined, raise an exception.
#
# If the passed in DN has more than one string '/CN=' raise an exception.
#
# If the passed in DN has no /CN= return undef.
#
# Allow whitespace around the "=".
sub dn_to_cn {
    my ($DN) = @_ ;

    if (!$DN) {
        croak 'passed in DN is empty or undefined' ;
    } elsif ($DN =~ m{\/CN\s*=\s*.*\/CN=}ixsm) {
        croak 'passed in DN has more than one CN' ;
    } elsif ($DN =~ m{CN\s*=\s*(.*?)([,\/]|$)}ixsm) {
        my $CN = $1 ;
        return $CN ;
    } else {
        return ;
    }
}

# Given a wallet name and type, read the wallet object into a string and
# return that string. Note that this command must be run in the context of
# valid Kerberos credentials.
#
# If wallet returns an error, raise an exception.
sub get_wallet {
    my ($name, $type) = @_ ;

    if (!$name) {
        croak 'missing wallet object name' ;
    }

    if (!$type) {
        $type = 'file' ;
    }

    my @cmd = ('wallet',
               'get',
               $type,
               $name,
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

    if ($stderr) {
        croak qq{failed to read wallet object "$type $name": $stderr} ;
    }

    return trim($stdout) ;
}

sub tail_file_matches {
    my ($filepath, $regex, $number_lines) = @_ ;

    # 1. Create a couple of temporary file.
    my $tfh          = File::Temp->new(UNLINK => 1) ;
    my $tailfile_tmp = $tfh->filename();

    # 2. Put the last $number_lines into $tempfile.
    my @cmd = (
        'tail',
        '-n', $number_lines,
        ) ;

    my $err ;

    IPC::Run::run(\@cmd, '<', $filepath, '>', $tailfile_tmp, '2>>', \$err) ;
    if ($err) {
        croak "error running tail: $err" ;
    }

    open (my $TAIL, '<', $tailfile_tmp) ;
    my @matching_lines = () ;
    while (my $line = <$TAIL>)
    {
        if ($line =~ $regex) {
            push(@matching_lines, trim($line)) ;
        }
    }
    close($TAIL) ;

    return @matching_lines ;
}

# Given a keytab file, create a ticket file and set KRB5CCNAME environment
# variable to point to this ticket file.
sub setup_kerberos_ticket {
    my ($keytab_file, $tgt_file) = @_ ;

    if (!$keytab_file) {
        croak 'missing keytab file' ;
    }

    if (!$tgt_file) {
        croak 'missing tgt file location' ;
    }
    my @cmd = (
        'k5start',
        '-U',
        '-f', $keytab_file,
        '-k', $tgt_file,
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command_improved(@cmd) ;

    if ($stderr) {
        croak "error creating TGT file from keytab file '$keytab_file': $stderr" ;
    } else {
        ## no critic (Variables::RequireLocalizedPunctuationVars)
        $ENV{'KRB5CCNAME'} = "FILE:$tgt_file" ;
        return 1 ;
    }
}


# Remove comments from POD mark-up from non-Perl documents
# and return as a string.
sub extract_pod {
    my ($filename) = @_ ;

    # STEP 1: get the content into a string.
    my $result = q{} ;
    if (!$filename) {
        # Read from STDIN
        ## no critic (InputOutput::ProhibitExplicitStdin);
        while (<STDIN>) {
            $result .= $_ ;
        }
    } else {
        $result = slurp($filename) ;
    }

    # ## ####    # ## ####    # ## ####    # ## ####    # ## ####
    my $is_pod_beginning_fref = sub {
        my ($l) = @_ ;
        if ($l =~ m{^[#]=pod}xsm) {
            return 1 ;
        } else {
            return 0 ;
        }
    } ;
    # ## ####    # ## ####    # ## ####    # ## ####    # ## ####
    my $is_pod_end_fref = sub {
        my ($l) = @_ ;
        if ($l =~ m{^[#]=cut}xsm) {
            return 1 ;
        } else {
            return 0 ;
        }
    } ;
    # ## ####    # ## ####    # ## ####    # ## ####    # ## ####
    my $strip_comment_fref = sub {
        my ($l) = @_ ;
        $l =~ s{^[#](.*)$}{$1}xsm ;
        return $l ;
    } ;
    # ## ####    # ## ####    # ## ####    # ## ####    # ## ####

    # Step 2. Find the POD start and stop strings and extract
    # the POD code.

    my $in_pod     = 0 ;
    my @pod_lines  = () ;
    my @lines      = split(/\n/xsm, $result) ;
    foreach my $line (@lines) {
        if ($is_pod_beginning_fref->($line)) {
            if ($in_pod) {
                croak 'parse error: already in a POD section' ;
            } else {
                $in_pod = 1 ;
                push(@pod_lines, $strip_comment_fref->($line)) ;
            }
        } elsif ($is_pod_end_fref->($line)) {
            if ($in_pod) {
                $in_pod      = 0 ;
                push(@pod_lines, $strip_comment_fref->($line)) ;
            } else {
                croak 'parse error: was not in a POD section' ;
            }
        } else {
            if ($in_pod) {
                push(@pod_lines, $strip_comment_fref->($line)) ;
            } else {
                # Nothing to do here.
            }
        }
    }

    if ($in_pod) {
        croak 'file ended before =cut' ;
    }

    return join("\n", @pod_lines) ;
}

# Prints man page to stdout.
sub print_man_page_from_pod {
    my ($filename) = @_ ;

    my $pod_text = extract_pod($filename) ;

    # Write the file
    my $tempfile = tempfile() ;

    my $header = <<'END_HEADER' ;
#!/usr/bin/perl
use strict ;
use warnings ;
use Pod::Usage ;
pod2usage(-verbose => 2) ;
exit 0 ;

END_HEADER

    open(my $MFH, '>', $tempfile) ;

    print {$MFH} $header ;
    print {$MFH} $pod_text ;

    close $MFH ;

    # The `perl $tempfile` will not work properly unless
    # the file readable by everyine.
    chmod 0644, $tempfile;

    ## no critic (InputOutput::ProhibitBacktickOperators);
    print `perl $tempfile` ;

    unlink $tempfile ;

    return ;
}

1 ;

__END__

=begin stopwords

tempfile config.txt abc xyz keytab
KRB5CCNAME MySQL unlink kerberos CN DN undef perldoc
Lewenberg

=end stopwords

=head1 NAME

Stanford::Orange::Util - Useful Perl functions

=head1 VERSION

Version 0.03

=head1 SYNOPSIS

    use Stanford::Orange::Util qw/trim tempfile slurp/ ;

    print trim(" xxx \n") ;
    # Result: prints "xxx" with no spaces or carriage return.

    my $temp_file = tempfile() ;
    spew($temp_file, "Hello, world.\n") ;
    # Result: a new file in the temporary directory with the contents
    # "Hello, world.\n".

    print(slurp($temp_file)) ;
    # Result: prints "Hello, world.\n".

=head1 DESCRIPTION

A collection of useful utility functions.

=head1 ABSTRACT

Contains several useful Perl functions including trim, slurp, spew,
tempfile, etc.

=head1 SUBROUTINES/METHODS

=head2 C<config_file_lines>

Converts a text file into an array with each line being an element of
the array. It remove any comments and any blank lines. All lines added
to the array are trimmed.

Example: If the file "config.txt" contains these five lines:

    # This is a config file.
    abc 123 xyz
      Hello, world!

    last line of the file

The result of C<config_file_lines('config.txt')> will be the array
    ('abc 123 xyz', 'Hello, world!', 'last line of the file')

=head2 C<files_older_than>

To do.


=head2 C<kerberos_ticket_cache>

Given a Kerberos keytab file and a file location, create a Kerberos ticket
cache in that file location. Sets the environment variable C<KRB5CCNAME> to
the location of the ticket cache

Example:

    kerberos_ticket_cache('/etc/keytab', '/tmp/krb.cache')

Creates a Kerberos ticket cache at F</tmp/krb.cache> and sets the
environment variable KRB5CCNAME equal to "/tmp/krb.cache". This function
attempts to provision an AFS token. If you do not want to provision an AFS
token use C<kerberos_ticket_cache_no_afs> instead


=head2 C<kerberos_ticket_cache_no_afs>

This function is exactly the same as C<kerberos_ticket_cache> except it
does not attempt to provision an AFS token.

=head2 C<kerberos_ticket_cache_aux>

The functions C<kerberos_ticket_cache> and C<kerberos_ticket_cache_no_afs>
are just wrapper functions around C<kerberos_ticket_cache_aux>. There
should be no need to ever call C<kerberos_ticket_cache_aux> directly.

=head2 C<mysql_time>

Returns the current time as a MySQL-compatible string:

    print mysql_time() . "\n" ;
    # Would print something like '2012-11-15 08:54:21'

=head2 C<number_recent_files>

Returns the number of recent files in a given directory

    print number_recent_files('/tmp', 60 * 60 * 24) ;
    # Would print the number of files with a last-modified date
    # in the last 24 hours.

=head2 C<slurp>

Returns the contents of a file. Be careful if the file is
large as C<slurp> will read the entire contents at once.

    $x = slurp('/etc/services') ;

The function takes an optional second argument to limit the file size
slurped. If the file is larger than this limit, the function will die.

    $x = slurp('/etc/services', 200) ;
    # This dies because '/etc/services' contains more than 200 bytes.

If the supplied is non-positive, it is silently ignored.

=head2 C<spew>

The opposite of C<slurp>, this function writes a string to a supplied
file. The file will be overwritten and not appended to. The
function returns 1 on success, and dies otherwise.

    spew('This is a string', '/tmp/newfile.txt') ;

=head2 C<tempfile>

(This function does not auto-unlink the file when the variable goes out of
scope, so it is recommended that you use it only when that is what you
want.)
Returns the full path to a new temporary file.
Creates a new temporary file using the C<File::Temp> module. Note that you
are responsible for deleting this temporary file; it will I<not> be
deleted when the scope is exited or the program finishes. The following
code will display "file exists":

    $filepath = tempfile() ;
    if (-e $filepath)
    {
      print "file exists\n" ;
    }

=head2 C<trim>

Given a scalar string returns the string with all leading and trailing
white space removed. If the argument passed is not defined returns C<undef>.
The following code will print C<YES>.

    if (trim(" x ") eq "x")
    {
      print "YES\n" ;
    }

=head2 C<run_command>

Returns standard output, standard error, and the exit code after running a
shell command. The following will print "cannot access /bogus: No such file or
directory":

    my ($stdout, $stderr, $exit_code) = run_command('ls /bogus') ;
    print $stderr . "\n" ;


=head2 C<run_command_improved>

Same as C<run_comand> but uses IPC::Run. Will probably replace C<run_command>.

=head2 C<run_under_keytab>

Given an array reference and a keytab file, run the command under the
kerberos context of that keytab file. Can be used to access AFS
file. Returns the same as C<run_command>.

    my ($stdout, $stderr, $exit_code)
      = run_command(['ls','/afs/ir/dir1'], '/etc/service.keytab') ;


=head2 C<dn_to_cn>

Extract the CN from a DN. Example:

    '/C=US/ST=California/L=Stanford/O=Stanford University/OU=IT Services MDM/CN=ios-device-test624.stanford.edu'

returns C<ios-device-test624.stanford.edu>.

If the passed in DN is empty or undefined, raise an exception. If the
passed in DN has more than one string C</CN=> raise an exception. If the
passed in DN has no C</CN=> return undef.


=head2 C<extract_pod>

Takes a file (or standard input) and extracts B<commented> POD code. For
example, if the file is:

    #!/usr/bin/python
    import re

    #=pod
    #
    #=head 1 First function
    #
    #This is the first function
    #
    #=cut

    def my_funct():
    ...

the string output will be

    =pod

    =head 1 First function

    This is the first function

    =cut

Note that you I<must> start and stop and every POD section with C<=pod> and C<=cut>.

You can take this output and run it through

=head2 C<print_man_page_from_pod>

Extracts POD from a file, formats it with Pod::Text and then prints it out.
To use this function, do this:


=head2 C<get_wallet>

Given a wallet name and type (defaults to 'file') returns the wallet
object's contents as a string. Note that this must be run in the context
of valid Kerberos credentials.

    my $config = get_wallet('config/my_server') ;
    # equivalent to
    my $config = get_wallet('config/my_server', 'file') ;

=head2 C<tail_file_matches>

Looks for matching lines at the end of a file.

    my @matched_lines = tail_file_matches($filepath, $regex, $number_lines) ;

All arguments are required. The last $number_lines of $filepath will be
examined and all those that match $regex will be returned.

=head2 C<setup_kerberos_ticket>

Given a keytab file, create a ticket file and set C<KRB5CCNAME>
environment variable to point to this ticket file.


=head1 DIAGNOSTICS

To be finished.

=head1 CONFIGURATION AND ENVIRONMENT

None needed.

=head1 DEPENDENCIES

Carp

=head1 INCOMPATIBILITIES

None known.

=head1 BUGS AND LIMITATIONS

None known.


=head1 AUTHOR

Adam Lewenberg, C<< <adamhl at stanford.edu> >>

=head1 BUGS

Please report any bugs or feature requests
to C<adamhl at stanford.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Stanford::Util

=head1 LICENSE AND COPYRIGHT

Copyright 2012 Stanford University, all rights reserved.

=cut
